/**
 * EasyBeans
 * Copyright (C) 2007-2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: LifeCycleCallback.java 2546 2008-02-29 13:49:31Z sauthieg $
 * --------------------------------------------------------------------------
 */
package org.ow2.easybeans.cxf;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.jws.WebService;

import org.apache.cxf.Bus;
import org.apache.cxf.BusException;
import org.apache.cxf.BusFactory;
import org.apache.cxf.jaxws.JaxWsServerFactoryBean;
import org.apache.cxf.jaxws.support.JaxWsImplementorInfo;
import org.apache.cxf.jaxws.support.JaxWsServiceFactoryBean;
import org.apache.cxf.transport.DestinationFactory;
import org.apache.cxf.transport.DestinationFactoryManager;
import org.ow2.easybeans.api.EZBContainerCallbackInfo;
import org.ow2.easybeans.api.Factory;
import org.ow2.easybeans.container.EmptyLifeCycleCallBack;
import org.ow2.easybeans.container.session.stateless.StatelessSessionFactory;
import org.ow2.easybeans.cxf.http.EasyBeansHTTPTransportFactory;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Callback for CXF.
 * @author Guillaume Sauthier
 * @author Florent Benoit
 */
public class LifeCycleCallback extends EmptyLifeCycleCallBack {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(LifeCycleCallback.class);

    /**
     * Easybeans internal Bus.
     */
    private Bus bus = null;

    /**
     * Initialize Internal CXF Bus.
     * @param info Callback Info.
     */
    private void initBus(final EZBContainerCallbackInfo info) {

        // Create a new Bus
        this.bus = BusFactory.newInstance().createBus();

        try {
            createDestinationFactory();
        } catch (BusException e) {
            logger.error("Cannot create destination directory", e);
        }

        logger.info("Internal EasyBeans Bus started ...");

    }

    /**
     * Create the EasyBeans destination factory and register the transports.
     * @throws BusException if the factory is not created.
     */
    private void createDestinationFactory() throws BusException {

        EasyBeansHTTPTransportFactory factory = new EasyBeansHTTPTransportFactory(this.bus);

        // transportIds done by Spring ?
        List<String> namespaces = new ArrayList<String>();
        namespaces.add("http://schemas.xmlsoap.org/wsdl/soap/http");
        namespaces.add("http://schemas.xmlsoap.org/soap/http");
        namespaces.add("http://www.w3.org/2003/05/soap/bindings/HTTP/");
        namespaces.add("http://schemas.xmlsoap.org/wsdl/http/");
        namespaces.add("http://cxf.apache.org/transports/http/configuration");
        namespaces.add("http://cxf.apache.org/bindings/xformat");
        factory.setTransportIds(namespaces);

        for (String ns : namespaces) {
            registerTransport(factory, ns);
        }

    }

    /**
     * Utility method to register TransportFactory under a namespace.
     * @param factory the TransportFactory instance
     * @param namespace the key
     * @throws BusException if registration fails.
     */
    private void registerTransport(final DestinationFactory factory, final String namespace) throws BusException {
        this.bus.getExtension(DestinationFactoryManager.class).registerDestinationFactory(namespace, factory);
    }

    /**
     * Called when container is starting.
     * @param info some information on the container which is starting.
     */
    @Override
	public void start(final EZBContainerCallbackInfo info) {

        initBus(info);

        // Iterates over the factories
        Map<String, Factory<?, ?>> factories = info.getFactories();
        for (Factory f : factories.values()) {

            // Only for @Stateless
            if (f instanceof StatelessSessionFactory) {

                StatelessSessionFactory factory = (StatelessSessionFactory) f;

                Class<?> klass = factory.getBeanClass();

                WebService ws = klass.getAnnotation(WebService.class);

                // only for @WebService
                if (ws != null) {

                    String serviceName = ws.serviceName();
                    String wsdlName = ws.wsdlLocation();
                    String portName = ws.portName();
                    String name = ws.name();

                    // Load the WSDL if possible
                    URL url = null;
                    if (wsdlName != null) {
                        url = klass.getClassLoader().getResource(wsdlName);
                        if (url == null) {
                            try {
                                url = new URL(wsdlName);
                            } catch (MalformedURLException ex2) {
                                url = klass.getClassLoader().getResource("/" + wsdlName);
                            }
                        }
                    }

                    JaxWsImplementorInfo jaxWsImplementorInfo = new JaxWsImplementorInfo(klass);
                    JaxWsServiceFactoryBean jaxWsServiceFactoryBean = new JaxWsServiceFactoryBean(jaxWsImplementorInfo);
                    jaxWsServiceFactoryBean.setBus(this.bus);
                    if (url != null) {
                        jaxWsServiceFactoryBean.setWsdlURL(url);
                    }

                    // Create a service with the correct classloader
                    ClassLoader old = Thread.currentThread().getContextClassLoader();
                    try {
                        Thread.currentThread().setContextClassLoader(klass.getClassLoader());
                        jaxWsServiceFactoryBean.create();
                    } finally {
                        // reset
                        Thread.currentThread().setContextClassLoader(old);
                    }

                    // Build the EasyBeans endpoint
                    JaxWsServerFactoryBean factoryBean = new EasyBeansJaxWsServerFactoryBean(jaxWsServiceFactoryBean, factory);
                    EasyBeansEndpoint endpoint = new EasyBeansEndpoint(factory, this.bus, factoryBean);

                    // find the URL pattern to use
                    String pattern = null;
                    PortComponent pc = klass.getAnnotation(PortComponent.class);
                    if (pc != null) {
                        pattern = pc.urlPattern();
                    } else {
                        if (portName != null) {
                            // default pattern : service-name/port-name
                            pattern = "/" + serviceName + "/" + portName;
                        } else {
                            pattern = "/" + serviceName + "/" + name;
                        }
                    }

                    // ... and finally publish using the URL pattern
                    endpoint.publish(pattern);

                    logger.info("Endpoint {0} published at {1}", klass.getName(), pattern);

                }
            }
        }
        logger.info("Bus started for " + info.getArchive());
    }

    /**
     * Called when container is stopping.
     * @param info some information on the container which is stopping.
    */
    @Override
	public void stop(final EZBContainerCallbackInfo info) {

        // Stop the Bus
        this.bus.shutdown(true);
        logger.info("Bus stopped for " + info.getArchive());

    }

}
