/**
 * EasyBeans
 * Copyright (C) 2007-2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ServletHelper.java 2546 2008-02-29 13:49:31Z sauthieg $
 * --------------------------------------------------------------------------
 */
package org.ow2.easybeans.cxf.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.Principal;
import java.util.Collection;
import java.util.Set;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.cxf.Bus;
import org.apache.cxf.common.util.StringUtils;
import org.apache.cxf.message.ExchangeImpl;
import org.apache.cxf.message.Message;
import org.apache.cxf.message.MessageImpl;
import org.apache.cxf.security.SecurityContext;
import org.apache.cxf.service.model.EndpointInfo;
import org.apache.cxf.transport.http.AbstractHTTPDestination;
import org.apache.cxf.transport.http.HTTPSession;
import org.apache.cxf.transport.http.WSDLQueryHandler;
import org.apache.cxf.transport.https.SSLUtils;
import org.apache.cxf.transport.servlet.ServletDestination;
import org.apache.cxf.transports.http.QueryHandler;
import org.apache.cxf.transports.http.QueryHandlerRegistry;
import org.ow2.easybeans.cxf.EasyBeansWSDLQueryHandler;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This class is almost the same code that comes from CXF.
 * {@link org.apache.cxf.transport.servlet.ServletController} CXF should provide
 * an helper class.
 * @author Florent Benoit
 */
public final class ServletHelper {

    /**
     * 404 error.
     */
    private static final int NOT_FOUND_ERROR = 404;

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ServletHelper.class);

    /**
     * An HTTP request.
     */
    private HttpServletRequest request;

    /**
     * An HTTP response.
     */
    private HttpServletResponse response;

    /**
     * An HTTP request.
     */
    private String lastBase = "";

    /**
     * The link to the {@link EasyBeansDestination}. (may be null)
     */
    private EasyBeansDestination destination;

    /**
     * CXF Bus.
     */
    private Bus bus;

    /**
     * Link to a Servlet Context.
     */
    private ServletContext servletContext;

    /**
     * Build a new helper class for the given data.
     * @param request the HTTP request
     * @param response the HTTP response
     * @param bus the CXF bus
     * @param servletContext the Servlet context
     */
    private ServletHelper(final HttpServletRequest request,
                          final HttpServletResponse response,
                          final EasyBeansDestination destination,
                          final ServletContext servletContext) {
        this.request = request;
        this.response = response;
        this.destination = destination;
        if (destination != null) {
            this.bus = destination.getBus();
        }
        this.servletContext = servletContext;
    }

    /**
     * Update the Endpoint addresses for all destinations.
     */
    private synchronized void updateDests() {
        String base = getBaseURL();

        if (base.equals(this.lastBase)) {
            return;
        }

        Set<String> paths = URLMapper.getInstance().getServantMap().keySet();
        for (String path : paths) {
            EasyBeansDestination destination = URLMapper.getInstance().getServantMap().get(getTrimmedPath(path));
            String address = destination.getEndpointInfo().getAddress();
            if (address.equals(path) || address.equals(this.lastBase + path)) {
                destination.getEndpointInfo().setAddress(base + path);
            }
        }
        this.lastBase = base;
    }

    /**
     * Invoke the destination by building a message from the given HTTP request.
     * @throws ServletException if invocation fails.
     */
    public void invoke() throws ServletException {
        try {

            if ("GET".equals(this.request.getMethod())) {
                updateDests();
            }

            if (this.destination == null) {
                if (this.request.getRequestURI().endsWith("services")
                    || this.request.getRequestURI().endsWith("services/")
                    || StringUtils.isEmpty(this.request.getPathInfo())
                    || "/".equals(this.request.getPathInfo())) {
                    generateServiceList();
                } else {
                    logger.warn("Can't find the the request for ''{0}''s Observer ", this.request.getRequestURL());
                    generateNotFound();
                }
            } else {
                EndpointInfo ei = this.destination.getEndpointInfo();
                if (null != this.request.getQueryString()
                    && this.request.getQueryString().length() > 0
                    && this.bus.getExtension(QueryHandlerRegistry.class) != null) {

                    String ctxUri = this.request.getPathInfo();
                    String baseUri = this.request.getRequestURL().toString() + "?" + this.request.getQueryString();

                    for (QueryHandler qh : this.bus.getExtension(QueryHandlerRegistry.class).getHandlers()) {
                        if (qh.isRecognizedQuery(baseUri, ctxUri, ei)) {

                            if (qh instanceof WSDLQueryHandler) {
                                qh = new EasyBeansWSDLQueryHandler(this.bus,
                                                                   this.request.getRequestURL().toString());
                            }

                            this.response.setContentType(qh.getResponseContentType(baseUri, ctxUri));
                            OutputStream out = this.response.getOutputStream();
                            try {
                                qh.writeResponse(baseUri, ctxUri, ei, out);
                                out.flush();
                                return;
                            } catch (Exception e) {
                                throw new ServletException(e);
                            }
                        }
                    }
                }

                invokeDestination(this.destination);
            }
        } catch (IOException e) {
            throw new ServletException(e);
        }
    }

    /**
     * Generates a list of services.
     * @throws IOException if the list cannot be built.
     */
    private void generateServiceList() throws IOException {
        Collection<EasyBeansDestination> destinations = URLMapper.getInstance().getServantMap().values();
        this.response.setContentType("text/html");
        this.response.getWriter().write("<html><body>");

        if (destinations.size() > 0) {
            for (ServletDestination sd : destinations) {
                if (null != sd.getEndpointInfo().getName()) {
                    String address = sd.getEndpointInfo().getAddress();
                    this.response.getWriter().write("<p> <a href=\"" + address + "?wsdl\">");
                    this.response.getWriter().write(sd.getEndpointInfo().getName() + "</a> </p>");
                }
            }
        } else {
            this.response.getWriter().write("No service was found.");
        }
        this.response.getWriter().write("</body></html>");
    }


    /**
     * Gets the base URL from the current request.
     * @return the base of the URL.
     */
    private String getBaseURL() {
        String reqPrefix = this.request.getRequestURL().toString();

        String pathInfo = "";
        if (this.request.getPathInfo() != null) {
            pathInfo = this.request.getPathInfo();
        }
        reqPrefix = reqPrefix.substring(0, reqPrefix.length() - pathInfo.length());
        return reqPrefix;
    }

    /**
     * Generates a 404 error for the given request.
     * @throws IOException if writing fails
     */
    protected void generateNotFound() throws IOException {
        this.response.setStatus(NOT_FOUND_ERROR);
        this.response.setContentType("text/html");
        this.response.getWriter().write("<html><body>No service was found.</body></html>");
    }

    /**
     * Invoke the destination object.
     * @param destination the given destination
     * @throws ServletException if the invocation fails.
     */
    public void invokeDestination(final EasyBeansDestination destination) throws ServletException {
        logger.debug("Receiving request on the thread ''{0}''", Thread.currentThread());

        try {
            MessageImpl inMessage = new MessageImpl();
            inMessage.setContent(InputStream.class, this.request.getInputStream());
            inMessage.put(AbstractHTTPDestination.HTTP_REQUEST, this.request);
            inMessage.put(AbstractHTTPDestination.HTTP_RESPONSE, this.response);
            inMessage.put(AbstractHTTPDestination.HTTP_CONTEXT, this.servletContext);
            inMessage.put(Message.HTTP_REQUEST_METHOD, this.request.getMethod());
            inMessage.put(Message.PATH_INFO, this.request.getPathInfo());
            inMessage.put(Message.QUERY_STRING, this.request.getQueryString());
            inMessage.put(Message.CONTENT_TYPE, this.request.getContentType());
            inMessage.put(Message.BASE_PATH, destination.getAddress().getAddress().getValue());
            inMessage.put(SecurityContext.class, new SecurityContext() {
                public Principal getUserPrincipal() {
                    return ServletHelper.this.request.getUserPrincipal();
                }

                public boolean isUserInRole(final String role) {
                    return ServletHelper.this.request.isUserInRole(role);
                }
            });

            // work around a bug with Jetty which results in the character
            // encoding not being trimmed correctly.
            String enc = this.request.getCharacterEncoding();
            if (enc != null && enc.endsWith("\"")) {
                enc = enc.substring(0, enc.length() - 1);
            }

            inMessage.put(Message.ENCODING, enc);
            SSLUtils.propogateSecureSession(this.request, inMessage);

            ExchangeImpl exchange = new ExchangeImpl();
            exchange.setInMessage(inMessage);
            exchange.setSession(new HTTPSession(this.request));

            destination.doMessage(inMessage);
        } catch (IOException e) {
            throw new ServletException(e);
        }

    }

    /**
     * Invokes the destination with the given parameters.
     * @param request the given HTTP request
     * @param response the HTTP response
     * @param destination the given transport factory
     * @param servletContext the Servlet context
     * @throws ServletException if object is not invoked.
     */
    public static void invoke(final HttpServletRequest request,
                              final HttpServletResponse response,
                              final EasyBeansDestination destination,
                              final ServletContext servletContext)
            throws ServletException {
        new ServletHelper(request, response, destination, servletContext).invoke();
    }

    /**
     * Remove the HTTP://localhost or https:/localhost prefix.
     * @param path the given path to handle
     * @return the path without the prefix
     */
    public static String getTrimmedPath(final String path) {
        String buildPath = path;
        if (buildPath == null) {
            return "/";
        }
        final String lh = "http://localhost/";
        final String lhs = "https://localhost/";

        if (buildPath.startsWith(lh)) {
            buildPath = buildPath.substring(lh.length());
        } else if (buildPath.startsWith(lhs)) {
            buildPath = buildPath.substring(lhs.length());
        }
        if (!buildPath.startsWith("/")) {
            buildPath = "/" + buildPath;

        }
        return buildPath;
    }
}
